
//
// Copyright (C) 2003-2005 Archie L. Cobbs and others.
// All rights reserved.
// 
// This file is part of the JC virtual machine software.
// 
// JC is free software; you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License,
// or (at your option) any later version.
// 
// JC is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
// Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with JC in the file "LGPL-2.1"; if not, write to:
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307, USA.
//
// $Id: JarLoader.java,v 1.1.1.1 2004/02/20 05:15:26 archiecobbs Exp $
//

package org.dellroad.jc.vm;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.jar.Attributes;
import java.util.jar.JarFile;
import java.util.jar.Manifest;

/**
 * Helper class used to implement the "--jar" command line flag.
 */
public class JarLoader {

	private JarLoader() {
	}

	/**
	 * Invoke the main class from a JAR file.
	 *
	 * @param args command line parameters with the JAR file
	 *	name prepended as the first element of the array.
	 */
	public static void main(String[] args) throws Throwable {

		// Open JAR file
		File jarfile = new File(args[0]);
		JarFile jar;
		try {
			jar = new JarFile(jarfile);
		} catch (IOException e) {
			throw new RuntimeException("can't open ``"
			    + jarfile + "'': " + e.toString());
		}

		// Find manifest and read main class attribute
		Manifest manifest;
		try {
			manifest = jar.getManifest();
		} catch (IOException e) {
			throw new RuntimeException("can't read manifest"
			    + " in ``" + jarfile + "'': " + e.toString());
		}
		if (manifest == null) {
			throw new RuntimeException("no manifest found in ``"
			    + jarfile + "''");
		}
		String main = manifest.getMainAttributes().getValue(
		    Attributes.Name.MAIN_CLASS);
		if (main == null) {
			throw new RuntimeException("no ``"
			    + Attributes.Name.MAIN_CLASS + "'' attribute"
			    + " found in manifest of ``" + jarfile + "''");
		}

		// Prepend JAR file to class path
		System.setProperty("java.class.path",
		    "file://" + jarfile.getAbsolutePath()
		      + System.getProperty("path.separator", ":")
		      + System.getProperty("java.class.path", "."));

		// Find the main class
		Class cl;
		try {
			cl = ClassLoader.getSystemClassLoader().loadClass(main);
		} catch (ClassNotFoundException e) {
			throw new RuntimeException("class ``" + main + "'' not"
			    + " found in JAR file ``" + jarfile + "''");
		}

		// Find the main() method
		Method method;
		try {
			method = cl.getDeclaredMethod("main",
			    new Class[] { String[].class });
		} catch (NoSuchMethodException e) {
			throw new RuntimeException("no main() method found"
			    + " in class ``" + main + "''");
		}

		// Invoke main() method
		String[] mainArgs = new String[args.length - 1];
		System.arraycopy(args, 1, mainArgs, 0, mainArgs.length);
		try {
			method.invoke(null, new Object[] { mainArgs });
		} catch (IllegalAccessException e) {
			throw new RuntimeException("can't invoke main() method"
			    + " in class ``" + main + "'': " + e.toString());
		} catch (InvocationTargetException e) {
			throw e.getTargetException();
		}
	}
}

