
/*
 * Copyright (C) 2003-2005 Archie L. Cobbs and others.
 * All rights reserved.
 * 
 * This file is part of the JC virtual machine software.
 * 
 * JC is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License,
 * or (at your option) any later version.
 * 
 * JC is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 * Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with JC in the file "LGPL-2.1"; if not, write to:
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * $Id: array.c,v 1.6 2005/03/13 00:18:51 archiecobbs Exp $
 */

#include "libjc.h"

/* Internal functions */
static jint	_jc_gen_array_hash_tables(_jc_env *env);

/*
 * Do one-time initialization for array types.
 *
 * If unsuccessful an exception is stored.
 */
jint
_jc_setup_array_types(_jc_env *env)
{
	jint status;

	/* Sanity check */
	_JC_ASSERT(env->vm->initialization != NULL);

	/* Generate array type's hash and "quick" lookup tables */
	if ((status = _jc_gen_array_hash_tables(env)) != JNI_OK)
		return status;

	/* Done */
	return JNI_OK;
}

/*
 * Generate interface list and hash tables for array types.
 *
 * If unsuccessful an exception is stored.
 */
static jint
_jc_gen_array_hash_tables(_jc_env *env)
{
	_jc_jvm *const vm = env->vm;
	_jc_boot_array *const array = &vm->boot.array;
	_jc_type *const interfaces[] = {	/* arrays implement these */
		vm->boot.types.Cloneable,
		vm->boot.types.Serializable,
	};
	const int num_interfaces = sizeof(interfaces) / sizeof(*interfaces);
	_jc_method *clone;
	int bucket;

	/* Sanity check */
	_JC_ASSERT(array->interfaces == NULL);
	_JC_ASSERT(array->imethod_hash_table == NULL);
	_JC_ASSERT(array->imethod_quick_table == NULL);

	/* Lock boot loader */
	_JC_MUTEX_LOCK(env, vm->boot.loader->mutex);

	/* Allocate interface list */
	if ((array->interfaces = _jc_cl_zalloc(env, vm->boot.loader,
	    num_interfaces * sizeof(*array->interfaces))) == NULL)
		goto fail;

	/* Fill in interface list */
	array->num_interfaces = num_interfaces;
	memcpy(array->interfaces, interfaces,
	    num_interfaces * sizeof(*interfaces));

	/* Allocate hash table and quick lookup table */
	if ((array->imethod_hash_table = _jc_cl_zalloc(env, vm->boot.loader,
	    _JC_IMETHOD_HASHSIZE * sizeof(*array->imethod_hash_table))) == NULL)
		goto fail;
	if ((array->imethod_quick_table = _jc_cl_zalloc(env, vm->boot.loader,
	      _JC_IMETHOD_HASHSIZE * sizeof(*array->imethod_quick_table)))
	      == NULL)
		goto fail;

	/* Find Object.clone() method */
	if ((clone = _jc_get_declared_method(env, vm->boot.types.Object,
	    "clone", "()Ljava/lang/Object;", _JC_ACC_STATIC, 0)) == NULL)
		goto fail;

	/* Get hash table index for clone() */
	bucket = clone->signature_hash & (_JC_IMETHOD_HASHSIZE - 1);

	/* Create hash table bucket */
	if ((array->imethod_hash_table[bucket] = _jc_cl_zalloc(env,
	    vm->boot.loader, 2 * sizeof(*array->imethod_hash_table[bucket])))
	    == NULL)
		goto fail;

	/* Put clone() in the bucket */
	array->imethod_hash_table[bucket][0] = clone;
	array->imethod_hash_table[bucket][1] = NULL;

	/*
	 * Put clone() in the "quick" method lookup table. It's the only
	 * interface method implemented by arrays so we trivially know
	 * there are no other methods in the same hash bucket.
	 */
	array->imethod_quick_table[bucket] = clone->function;

	/* Unlock boot loader */
	_JC_MUTEX_UNLOCK(env, vm->boot.loader->mutex);

	/* Done */
	return JNI_OK;

fail:
	/* Clean up after failure */
	_jc_cl_unalloc(vm->boot.loader, &array->imethod_quick_table,
	    _JC_IMETHOD_HASHSIZE * sizeof(*array->imethod_quick_table));
	_jc_cl_unalloc(vm->boot.loader, &array->imethod_hash_table,
	    _JC_IMETHOD_HASHSIZE * sizeof(*array->imethod_hash_table));
	_jc_cl_unalloc(vm->boot.loader, &array->interfaces,
	    num_interfaces * sizeof(*array->interfaces));
	_JC_MUTEX_UNLOCK(env, vm->boot.loader->mutex);
	return JNI_ERR;
}


